/*------------------------------------------------------------------------------
    GALE. Fine-grained parallel Genetic Algorithms for Data Mining.
    Copyright (C) 2002  Xavier Llor <xllora@illigal.ge.uiuc.edu>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
------------------------------------------------------------------------------*/


package edu.uiuc.ge.illigal.gale;

import edu.uiuc.ge.illigal.gale.config.*;
import edu.uiuc.ge.illigal.gale.data.*;
import edu.uiuc.ge.illigal.gale.agents.*;
import edu.uiuc.ge.illigal.gale.board.*;
import edu.uiuc.ge.illigal.gale.io.*;

import java.util.Date;
import java.util.Vector;

/**
 * Entry point for running GALE environment.
 *
 * @author  Xavier Llor&agrave; &lt;<a href="mailto:xllora@illigal.ge.uiuc.edu">xllora@illigal.ge.uiuc.edu</a>&gt;
 * @version 0.9alpha
 * @since 0.9alpha
 */

public class GALE
{
	//-----------------------------------------------------------------------

	/** GALE version identification sting */
	public final static String sVersion = "gale.0.9.alpha-release";

	//-----------------------------------------------------------------------

	/** Configuration file name */
	protected String sGALEConfigFile;

	/** Train file name */
	protected String sTrainFile;

	/** Test file name (if provided, cross-validation disabled) */
	protected String sTestFile;

	/** Cross validation fold size used */
	protected boolean bFolds;

	/** Number of folds */
	protected int iKFolds;

	/** GALE configuration object */
	protected GALEConfiguration galeCnf;

	/** GALE training instance set */
	protected InstanceSet insSet;

	/** GALE test instance set */
	protected InstanceSet insTestSet;

	/** The board used by GALE */
	protected Board brd;

	/** The IO Factory */
	protected IOFactory ioFac;

	/** Save folders required */
	protected boolean bSaveFiles;

	/** Hold out requires */
	protected boolean bHoldOut;

	/** Use hold out */
	protected float fHoldOutProp;

	/** Data of startup */
	protected Date dateStart;

	/** Board dump flag */
	protected boolean bDump;

	//-----------------------------------------------------------------------

	/**
    * Configures GALE environment
	 *
	 * @param sArgs command line arguments
	 * @throws Exception could not initialize GALE environment
    */
	public GALE ( String [] sArgs )
	throws Exception
	{
		// Update the configuration from the command line
		updateOptions(sArgs);

		// Loads the GALE configuration file
		if ( sGALEConfigFile!=null)
			galeCnf = new FileGALEConfiguration(sGALEConfigFile);
		else
			throw new Exception("Configuration file (-c option) required");

		// Load the training file
		if ( sTrainFile!=null )
			insSet = new FileInstanceSet(galeCnf, sTrainFile);
		else
			throw new Exception("Training file (-t option) required" );

		// Initializing distance function
		galeCnf.getDistance().init(insSet.attributeSet());
		
		// Load the test file (if any)
		if ( sTestFile!=null )
			insTestSet = new FileInstanceSet(galeCnf, sTestFile);
		else 
			insTestSet = null;

		// Check for cross-validation and holdout
		if ( bFolds && bHoldOut )
			throw new Exception("Could not apply cross-validation and holdout at the same time");

		// Check for inconsistencies
		if ( sTestFile!=null && bSaveFiles )
			throw new Exception("Save option not allowed when test file provided");

		ioFac = new FileIOFactory ( 100 );
	
		dateStart = new Date();
		ioFac.openChannel("stdout","GALE."+insSet.getRelationName()+"."+dateStart.getTime()+".stdout"); 
	}

	//-----------------------------------------------------------------------

	/**
    * Runs GALE environment.
	 *
	 * @throws Exception could not initialize GALE environment
    */
	public void run ()
	throws Exception
	{
		ioFac.push("stdout", "***************************"+ 
		                     "****************************");
		ioFac.push("stdout", "****** GALE environment ("+sVersion+")  ******");
		ioFac.push("stdout", "***************************"+
		                     "****************************\n");

		ioFac.push("stdout", "Started at : "+dateStart+"\n" );

		if ( bHoldOut || sTestFile!=null )
			trainTestRun();
		else
			crossValidation ();

		ioFac.push("stdout", "\nEnded at : "+new Date() );
	}

	//-----------------------------------------------------------------------

	/**
    * Runs cross-validation.
	 *
	 * @throws Exception could not initialize GALE environment
    */
	protected void crossValidation()
	throws Exception
	{ 
		AgentPerformance [] apa = new AgentPerformance[iKFolds];

		float fMean = 0;
		float fDev  = 0;
		float fTmp  = 0;

		// Identify what's going on
		ioFac.push("stdout", "****** K-Fold Stratified Cross-validation Run ******\n");

		// Build the stratified folds for k-cross-validation
		OrderedSet [] osaFolds = FoldGenerator.stratifiedFolds(
		                                      insSet,
		                                      iKFolds,
		                                      galeCnf.getNumClasses(),
		                                      galeCnf.getRandom()
		                                     );

		// Save folds (if required)
		if ( bSaveFiles ) {
			ioFac.push("stdout", "-s option found.\n\nSaving the "+iKFolds+
			                     " generated folds for "+insSet.getSize()+"...\n");
			saveFolds(osaFolds);
			return;
		}

		// Builds the board
		galeCnf.getRandom().setSeed(galeCnf.getSeedRuns());
		brd = new Board ( galeCnf, insSet, osaFolds, ioFac, bDump );

		// Obtain a full model
		brd.runFold(-1,insSet.getRelationName()+"."+dateStart.getTime());
		ioFac.push("stdout", brd.bestAgent());
		
		// Identify what's going on
		ioFac.push("stdout", "\n****** Cross-validation with "+iKFolds+" folds ******\n");

		// Cross-validation run
		for ( int i=0 ; i<iKFolds ; i++ ) {
			// Run the fold
			brd.runFold(i,insSet.getRelationName()+"."+dateStart.getTime());
			// Run the test
			switch ( galeCnf.getTestMode() ) {
				case GALEConfiguration.TEST_BEST:
				                     apa[i] = bestAgentTest(osaFolds[i]); break;
				case GALEConfiguration.TEST_BAGGING:
			   	                  apa[i] = baggingTest(osaFolds[i]); break;
				default:
					throw new Exception ( galeCnf.getTestMode() +
				                      "not valid test mode constant" );
			}
			// -------------------- CV summary -------------------------------
			int iTotal    = apa[i].getTotal();
			int iOK       = (int)(apa[i].getAccuracy()*iTotal);
			int iDontKnow = apa[i].getDontKnow();
			int iNotOK    = iTotal-iOK-iDontKnow;
			ioFac.push("stdout", "CV fold "+i+" :"
			              +"\t"+iOK
			              +"\t"+iNotOK
			              +"\t"+iDontKnow
			              +"\t"+iTotal
			              +"\t"+100.0f*apa[i].getAccuracy()
			              +"\t"+100.0f*((float)iOK/(iOK+iNotOK)) );

			// -------------------- Save test results ------------------------
			String sFold = ""+i;
			for ( int k=6, j=sFold.length() ; k>j ; k-- )
				sFold = "0"+sFold;
			sFold = "f"+sFold;
			// Initializing the channels	
			ioFac.openChannel("test","GALE."+insSet.getRelationName()+"."+dateStart.getTime()
	 		                         +".board."+sFold+".test"); 
			ioFac.push("test",apa[i]);
			ioFac.closeChannel("test");
			// ---------------------------------------------------------------


		}
		
		// Compute the final performance
		for ( int i=0 ; i<iKFolds ; i++ ) 
			fMean += apa[i].getAccuracy();
		fMean /= iKFolds;

		for ( int i=0 ; i<iKFolds ; i++ ) {
			fTmp  = fMean-apa[i].getAccuracy();
			fDev += fTmp*fTmp;
		}
		fDev  = (float)Math.sqrt(fDev/iKFolds);
	
		ioFac.push("stdout", "\n");
		ioFac.push("stdout", "Mean            : "+100*fMean);
		ioFac.push("stdout", "Deviation (com) : "+100*fDev);

	}

	//-----------------------------------------------------------------------

	/**
    * Runs train an test.
	 *
	 * @throws Exception could not initialize GALE environment
    */
	protected void trainTestRun()
	throws Exception
	{ 
		AgentPerformance ap = null;
		InstanceSet [] insa = new InstanceSet[2];

		// Generate the hold out sets
		if ( bHoldOut ) {
			// Identify what's going on
			ioFac.push("stdout", "****** Stratified holdout training run with : "+(100*fHoldOutProp)+"% of "+sTrainFile+" ******\n");

			insa = HoldOutGenerator.stratifiedHoldOut(
			                                insSet,
			                                fHoldOutProp,
			                                galeCnf
			                             );
		}
		else {
			// Identify what's going on
			ioFac.push("stdout", "****** Train run with : "+sTrainFile+" ******\n");

			insa[0] = insSet;
			insa[1] = insTestSet;
		}

		// Save folds (if required)
		if ( bSaveFiles ) {
			ioFac.push("stdout", "-s option found.\n\nSaving the hold out "+
			                     "files ...\n");
			saveHoldOut(insa);
			return;
		}

		// run the train
		galeCnf.getRandom().setSeed(galeCnf.getSeedRuns());
		brd = new Board ( galeCnf, insa[0], null, ioFac, bDump );
		brd.runFold(-1,insSet.getRelationName()+"."+dateStart.getTime());
		ioFac.push("stdout", brd.bestAgent());

		if ( bHoldOut ) {
			ioFac.push("stdout", "****** Stratified holdout test run with : "+(100*(1-fHoldOutProp))+"% of "+sTrainFile+" ******\n");
		}
		else {
			// Identify what's going on
			ioFac.push("stdout", "****** Test run with : "+sTrainFile+" ******\n");
		}

		// run the test
		switch ( galeCnf.getTestMode() ) {
			case GALEConfiguration.TEST_BEST:
			                     ap = bestAgentTest(insa[1]); break;
			case GALEConfiguration.TEST_BAGGING:
			                     ap = baggingTest(insa[1]); break;
			default:
				throw new Exception ( galeCnf.getTestMode() +
				                      "not valid test mode constant" );
		}
		ioFac.push("stdout", ap);
	}

	//-----------------------------------------------------------------------

	/**
    * Saves the holdout instance sets to files.
	 *
	 * @param insa the array of instance sets
	 * @throws Exception problems using the channels
    */
	protected void saveHoldOut ( InstanceSet [] insa )
	throws Exception
	{
			String sFile = null;

			//--------------- Train file ---------------

			// Opening the channel
			sFile = insa[0].getRelationName()+".arff";
			ioFac.openChannel(sFile,sFile);

			// Trace
			ioFac.push("stdout","Train file : "+sFile);

			// Write
			ioFac.push(sFile,insa[0]);

			// Closing the channel
			ioFac.closeChannel(sFile);
		
			//--------------- Test file ---------------

			// Opening the channel
			sFile = insa[1].getRelationName()+".arff";
			ioFac.openChannel(sFile,sFile);

			// Trace
			ioFac.push("stdout","Test file  : "+sFile);

			// Write
			ioFac.push(sFile,insa[1]);

			// Closing the channel
			ioFac.closeChannel(sFile);
	}


	//-----------------------------------------------------------------------

	/**
    * Tests using the best obtained agent.
	 *
	 * @param os the set of instances to use
	 * @throws Exception could not initialize GALE environment
    */
	protected AgentPerformance bestAgentTest ( OrderedSet os )
	throws Exception
	{
		Agent agn = (Agent)brd.bestAgent().clone();
	
		agn.resetPerformance();
		for ( int i=0, iMax=os.getSize() ; i<iMax ; i++ )
			agn.classify(os.getInstance(i));

		return agn.getPerformance();
	}

	//-----------------------------------------------------------------------

	/**
    * Tests using bagging.
	 *
	 * @param os the set of instances to use
	 * @throws Exception could not initialize GALE environment
    */
	protected AgentPerformance baggingTest ( OrderedSet os )
	throws Exception
	{
		AgentPerformance ap   = new BatchPerformance(galeCnf);
		Agent []         aa   = brd.fullMappedAgents();
		Instance         ins  = null;
		int              i    = 0;
		int              j    = 0;
		int              iMax = 0;
		int              iPos = 0;
		int              iCls = galeCnf.getNumClasses();
		int []           iap  = new int[iCls+1];
	
		for ( int iNumIns=os.getSize() ; i<iNumIns ; i++ ) {
			// Get the instance
			ins = os.getInstance(i);
			// Clean prediction array
			for ( j=0 ; j<iap.length ; j++ ) 
				iap[j]=0;  
			// Predict the class
			for ( j=0 ; j<aa.length ; j++ )
				iap[aa[j].classify(ins)+1]++;
				
			// Decide the final class
			for ( iMax=iap[0], iPos=0, j=0 ; j<iap.length ; j++ )
				if ( iMax<iap[j] ) {
					iMax = iap[j];
					iPos = j;
				}
	
			// Classify the instance into the most voted class
			ap.update(iPos-1,ins.getInstanceClass());
		}
		
		return ap;
	}
	
	//-----------------------------------------------------------------------

	/**
    * Updates the command line options for GALE environment.
	 *
	 * @param sArgs command line arguments
	 * @throws Exception command line error
    */
	protected void updateOptions ( String [] sArgs )
	throws Exception
	{
		int i=0, iMax=sArgs.length; 

		sTrainFile = sTestFile = sGALEConfigFile = null;
		iKFolds      = 10;
		bSaveFiles   = false;
		fHoldOutProp = -1;
		bFolds       = false;
		bHoldOut     = false;
		bDump        = false;
		
		while (i<iMax) {
			if (sArgs[i].equals("-t")) {
				if (i+1<iMax) sTrainFile = sArgs[++i];
				else throw new Exception("File name required for -t option");
			} 
			else if (sArgs[i].equals("-T")) {
				if (i+1<iMax) sTestFile = sArgs[++i];
				else throw new Exception("File name required for -T option");
			} 
			else if (sArgs[i].equals("-c")) {
				if (i+1<iMax) sGALEConfigFile = sArgs[++i];
				else throw new Exception("File name required for -c option");
			} 
			else if (sArgs[i].equals("-x")) {
				bFolds = true;
				if (i+1<iMax) iKFolds = Integer.parseInt(sArgs[++i]);
				else throw new Exception("Number of folds required for -x option");
				if ( iKFolds<2 ) 
					throw new Exception("Minimum number of fold for -x option is 2");
			} 
			else if (sArgs[i].equals("-s")) {
				bSaveFiles = true;
			} 
			else if (sArgs[i].equals("-h")) {
				bHoldOut = true;
				if (i+1<iMax) fHoldOutProp = Float.parseFloat(sArgs[++i]);
				else throw new Exception("Proportion required for -h option");
				if ( fHoldOutProp<=0 || fHoldOutProp>=1 )
					throw new Exception("Proportion for -h option must belong to (0,1) domain");
			} 
			else if (sArgs[i].equals("-b")) {
				bDump = true;
			}
			else throw new Exception(sArgs[i]+" unknown command line option");

			i++;
		}
	}

	//-----------------------------------------------------------------------

	/**
    * Saves the folds to files.
	 *
	 * @param osa the array of folds
	 * @throws Exception problems using the channels
    */
	protected void saveFolds ( OrderedSet [] osa )
	throws Exception
	{
		String  sFile = null;
		String   sTmp = null;
		OrderedSet os = null;

		for ( int i=0, iMax=osa.length ; i<iMax ; i++ ) {
			//------------------------------------------
			// Train fold dump
			// Building the extension
			sFile = "GALE."+insSet.getRelationName()+".cv.seed."+galeCnf.getSeed()+".fold."; 
			sTmp  = ""+i;
			for ( int j=6, l=sTmp.length() ; j>l ; j-- )
				sFile += '0';
			sFile+=sTmp+".train.arff";

			// Opening the channel
			ioFac.openChannel(sFile,sFile);

			// Building the fold
			os = new CVView(osa,i);
			
			// Trace
			ioFac.push("stdout",sFile+"\t"+os.getSize()+" instances");

			// Dumping the fold 
			dumpHeader(sFile,os.getSize());
			
			// Dumping instances
			for ( int j=0, iNI=os.getSize() ; j<iNI ; j++ ) 
				ioFac.push(sFile, os.getInstance(j));

			// Closing the channel
			ioFac.closeChannel(sFile);
		
			//------------------------------------------
			// Test fold dump
			// Building the extension
			sFile = "GALE."+insSet.getRelationName()+".cv.seed."+galeCnf.getSeed()+".fold."; 
			sTmp  = ""+i;
			for ( int j=6, l=sTmp.length() ; j>l ; j-- )
				sFile += '0';
			sFile+=sTmp+".test.arff";

			// Opening the channel
			ioFac.openChannel(sFile,sFile);

			// Building the fold
			os = osa[i];
			
			// Trace
			ioFac.push("stdout",sFile+"\t"+os.getSize()+" instances");

			// Dumping the fold 
			dumpHeader(sFile,os.getSize());
			
			// Dumping instances
			for ( int j=0, iNI=os.getSize() ; j<iNI ; j++ ) 
				ioFac.push(sFile, os.getInstance(j));

			// Closing the channel
			ioFac.closeChannel(sFile);

			//------------------------------------------
		}
	}
	

	/**
    * Dumps the header into the given channel
	 *
	 * @param chn the channel to use
	 * @param iNumIns the number of instances
	 * @throws Exception problems closing the channels
    */
	protected void dumpHeader ( String chn, int iNumIns )
	throws Exception
	{
		AttributeSet att = insSet.attributeSet();
		String    sClass = null;

		ioFac.push(chn, "%");
		ioFac.push(chn, "% GALE automatic stratified fold generation for CV");
		ioFac.push(chn, "%");
		ioFac.push(chn, "% Generated at        : "+new Date());
		ioFac.push(chn, "% Random seed         : "+galeCnf.getSeed());
		ioFac.push(chn, "% File name           : "+chn);
		ioFac.push(chn, "% Number of instances : "+iNumIns);
		ioFac.push(chn, "%\n");

		ioFac.push(chn, "@RELATION "+insSet.getRelationName()+"\n");

		ioFac.push(chn, att+"\n@DATA");

	}
	
	//-----------------------------------------------------------------------

	/**
    * Invocation method for running GALE environment.
    *
	 * @throws Exception problems closing the channels
    */
	public void close()
	throws Exception
	{
		ioFac.closeAllChannels();
	}

	//-----------------------------------------------------------------------

	/**
    * Returns the set of available option
	 *
	 * @return the options
    */
	protected static String listOptions()
	{
		String sRes = "";

		sRes += "****** GALE environment ("+sVersion+") ******\n\n";
		sRes += "Started at : "+ new Date()+"\n";
		sRes += "\nOptions required!!!\n";

		sRes += "----------------------------------------------------------------\n";
		sRes += "Options:\n";
		sRes += "-t <file_name> (required)\n";
		sRes += "\tSets the file name of the training data in arff format\n";

		sRes += "-c <file_name> (required)\n";
		sRes += "\tSets the configuration file for the GALE environment\n";

		sRes += "-x <folds> (optional)\n";
		sRes += "\tSets the number of folds for cross-validation\n";

		sRes += "-h <train_proportion> (optional)\n";
		sRes += "\tSave the train and test files for hold-out\n";

		sRes += "-s (optional)\n";
		sRes += "\tSave the files used \n";

		sRes += "-b (optional)\n";
		sRes += "\tSave the evolution of the board across the run\n";

		sRes += "-T <file_name> (optional)\n";
		sRes += "\tDisables cross-validation. Test the results using the arff file\n";

		sRes += "----------------------------------------------------------------\n";

		sRes += "\nEnded at : "+ new Date();

		return sRes;
	}

	//-----------------------------------------------------------------------

	/**
    * Invocation method for running GALE environment.
	 *
	 * @param sArgs command line arguments
    */
	public static void main ( String [] sArgs )
	{
		if ( sArgs.length==0 )
			System.out.println(listOptions());
		else {
			try {
				GALE gale = new GALE(sArgs);
			
				gale.run();
				gale.close();
			}
			catch ( Exception e ) {
				e.printStackTrace();
				//System.err.println(e);
			}
		}
	}

	//-----------------------------------------------------------------------
	
}
