/*------------------------------------------------------------------------------
    GALE. Fine-grained parallel Genetic Algorithms for Data Mining.
    Copyright (C) 2002  Xavier Llor <xllora@illigal.ge.uiuc.edu>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
------------------------------------------------------------------------------*/


package edu.uiuc.ge.illigal.gale.agents;

import edu.uiuc.ge.illigal.gale.config.*;
import edu.uiuc.ge.illigal.gale.data.*;

import java.io.Serializable;
import java.util.Vector;
import java.util.Random;

/** 
 * Implements the generic handle to Agents for Rule based agents.
 *
 * @author  Xavier Llor&agrave; &lt;<a href="mailto:xllora@illigal.ge.uiuc.edu">xllora@illigal.ge.uiuc.edu</a>&gt;
 * @version 0.9alpha 
 * @since   0.9alpha 
 *
 */

public class CLAgent
implements Agent, Serializable, Cloneable
{
	//-------------------------------------------------------------------------

	/** The GALE environment configuration object */
   protected GALEConfiguration galeCnf;

	/** The attributes information for instances */
   protected AttributeSet attCnf;

   /** Agent Performance object*/
	protected AgentPerformance agnPer;

	/** Random number generator object */
	protected Random rnd;

	/** Genetic material vector */
	protected Vector genVec;

	/** Prune information */
	protected int [] iaPrune;

	/** Match information */
	protected int [] iaMatch;

	//------------------------------------------------------------------------

   /** 
	 * Builds a simple agent with null information
    */ 
      
   public CLAgent ()
   {
		attCnf  = null;
   	galeCnf = null;
		agnPer  = null;
		rnd     = null;
		genVec  = null;
		iaPrune = null;
		iaMatch = null;
   }
   
	//------------------------------------------------------------------------

   /** 
	 * Builds a simple agent. The Agent is build using the GALE configuration
    * object.
	 *
    * @param gale The GALE configuration
	 * @param insSet The attribute set information used
    */ 
      
   public CLAgent ( GALEConfiguration gale, AttributeSet att )
   {
		attCnf  = att;
   	galeCnf = gale;
		rnd     = gale.getRandom();
		agnPer  = new BatchPerformance(galeCnf);

		genVec  = new Vector();

		// Genetic Material initialization
		int [] iaTmp = null;
		int iAtts = gale.getNumAttributes();
		for ( int i=0, iMax=gale.getAgentMaxInitWidth() ; i<iMax ; i++ ) {
			iaTmp = new int[iAtts+1];
			for ( int j=0 ; j<iAtts ; j++ )
				if ( rnd.nextBoolean() )
					iaTmp[j] = 0;
				else
					iaTmp[j] = rnd.nextInt((int)att.attribute(j).upperBound())+1;
			iaTmp[iAtts] = rnd.nextInt(galeCnf.getNumClasses());
			genVec.addElement(iaTmp);
		}
		
		// Purge information
		iaPrune = new int[genVec.size()];
		iaMatch = new int[genVec.size()];
   }
   
	//---------------------------------------------------------------------
	
	// Agent implementation
   public Agent buildAgent ( GALEConfiguration galeCnf, 
	                          AttributeSet attSet )
	{
		return new CLAgent ( galeCnf, attSet );
	}

	//-------------------------------------------------------------------------

	// Agent implementation
	public void resetPerformance ()
	{
		agnPer.reset();
		for ( int i=0 ; i<iaPrune.length ; i++ ) {
			iaPrune[i] = 0;
			iaMatch[i] = 0;
		}
	}

	//-------------------------------------------------------------------------

	// Agent implementation
	public AgentPerformance getPerformance ()
	{
		return agnPer;
	}


	//-------------------------------------------------------------------------

	// Agent implementation
	public Agent merge ( Agent agn )
	{
		CLAgent agnRes = null;
		CLAgent  agnIn = (CLAgent)agn;
		int []      ia = null;

		// Cut points for attributes
		int iAtts = galeCnf.getNumAttributes();
		int iCAP1 = rnd.nextInt(iAtts+1);
		int iCAP2 = rnd.nextInt(iAtts+1);

		// Computing the size of the agents
		int iST = genVec.size();
		int iSA = agnIn.genVec.size();

		// Cut points for the instances of this agent
		int iCITP1 = rnd.nextInt(iST);
		int iCITP2 = rnd.nextInt(iST);
		if ( iCITP2<iCITP1 ) {
			int iTmp = iCITP1;
			iCITP1   = iCITP2;
			iCITP2   = iTmp;
		}

		// Cut points for the instances of the input agent
		int iCIAP1 = rnd.nextInt(iSA);
		int iCIAP2 = rnd.nextInt(iSA);
		if ( iCIAP2<iCIAP1 ) {
			int iTmp = iCIAP1;
			iCIAP1   = iCIAP2;
			iCIAP2   = iTmp;
		}

		// Check if attribute cut points must be swapped
		if ( iCITP1==iCITP2 || iCIAP1==iCIAP2 ) 
			if ( iCAP2<iCAP1 ) {
				int iTmp = iCAP1;
				iCAP1    = iCAP2;
				iCAP2    = iTmp;
			}

		// Prepare the crossover structures
		int    iGenLen = iAtts+1;
		int iMaxLength = iST*iGenLen+iSA*iGenLen;
		int []   iaT   = new int[iST*iGenLen];
		int []   iaA   = new int[iSA*iGenLen];
		int []   iaRT  = new int[iMaxLength];
		int []   iaRA  = new int[iMaxLength];

		for ( int iArr=0, i=0 ; i<iST ; i++ ) {
			ia = (int[])genVec.elementAt(i);
			for ( int j=0 ; j<iGenLen ; j++, iArr++ )
				iaT[iArr] = ia[j];
		}

		for ( int iArr=0, i=0 ; i<iSA ; i++ ) {
			ia = (int[])agnIn.genVec.elementAt(i);
			for ( int j=0 ; j<iGenLen ; j++, iArr++ )
				iaA[iArr] = ia[j];
		}

		// Exchange the material
		int iXT1 = iCITP1*iGenLen+iCAP1;
		int iXT2 = iCITP2*iGenLen+iCAP2;
		int iXA1 = iCIAP1*iGenLen+iCAP1;
		int iXA2 = iCIAP2*iGenLen+iCAP2;

		int iT  = 0;
		int iA  = 0;
		int iRT = 0;
		int iRA = 0;

		// Coping the head of the genes
		for ( ; iT<iXT1 ; iT++, iRT++ ) iaRT[iRT] = iaT[iT]; 	
		for ( ; iA<iXA1 ; iA++, iRA++ ) iaRA[iRA] = iaA[iA]; 	

		// Swapping genes
		for ( ; iA<iXA2 ; iA++, iRT++ ) iaRT[iRT] = iaA[iA];
		for ( ; iT<iXT2 ; iT++, iRA++ ) iaRA[iRA] = iaT[iT];
		
		// Coping the tail of the genes
		for ( ; iT<iST*iGenLen ; iT++, iRT++ ) iaRT[iRT] = iaT[iT]; 	
		for ( ; iA<iSA*iGenLen ; iA++, iRA++ ) iaRA[iRA] = iaA[iA]; 	
		

		// Interpreting the resultant genetic material
		Vector vec = null;

		if ( iRT== 0 )
			vec = translateArray(iaRA,iRA);
		else if ( iRA== 0 )
			vec = translateArray(iaRT,iRT);
		else
			vec = (rnd.nextBoolean()) ? translateArray(iaRT,iRT)
		   	                       : translateArray(iaRA,iRA);
		
		// Remove the repeated rules
		removeRepeatedRules(vec);

		try {
			agnRes         = (CLAgent)clone();
			agnRes.genVec  = vec;
			agnRes.iaPrune = new int[vec.size()];
			agnRes.iaMatch = new int[vec.size()];
		}
		catch ( Exception e ) {
			System.err.println("Fail in merge for "+getClass().getName());
			e.printStackTrace();
			System.exit(1);
		}

		return agnRes;
	}

	//-------------------------------------------------------------------------

	/**
    * Remove the repeated rules.
	 *
	 * @param vec the vector to process
	 */
	protected void removeRepeatedRules ( Vector vec )
	{
		int        i = 0;
		int        j = 0;
		int []    ia = null;
		int [] iaTmp = null;

		for ( i=0 ; i<vec.size() ; i++ ) {
			ia = (int[])vec.elementAt(i);	
			for ( j=i+1 ; j<vec.size() ; j++ ) 
				if ( equalRules(ia,(int[])vec.elementAt(j)) ) {
					vec.removeElementAt(j);
					j--;
				}
		}
	}

	//-------------------------------------------------------------------------

	/**
    * Check if the rules are equal.
	 *
	 * @param ia1 the first rule
	 * @param ia2 the second rule
	 * @return true if the two rules are equal
	 */
	protected boolean equalRules ( int [] ia1, int [] ia2 )
	{
		boolean bRes = true;

		for ( int i=0, iMax=ia1.length ; bRes && i<iMax ; i++ )
			if ( ia1[i]!=ia2[i] )
				bRes = false;

		return bRes;
	}

	//-------------------------------------------------------------------------

	/**
    * Build the vector representation of a genetic array.
	 *
	 * @param iaGen the float array to translate
	 * @param iLen the length of useful information in the array
	 * @return the vector
	 */
	protected Vector translateArray ( int [] iaGen, int iLen )
	{
		int   iAtts = galeCnf.getNumAttributes();
		int   iMaxs = iAtts+1;
		int   iRows = iLen/iMaxs;
		Vector  vec = new Vector();
		int  []  ia = null;
		int    iAcc = 0;

		if ( iLen%iMaxs != 0 ) {
			System.err.println("Mistaken size after crossover "+iLen+","+iMaxs);
			System.exit(1);
		}
		
		for ( int i=0 ; i<iRows ; i++ ) {
			ia = new int[iMaxs];
			for ( int j=0 ; j<iMaxs ; j++, iAcc++ ) {
				ia[j] = iaGen[iAcc];
			}
			vec.addElement(ia);
		}

		return vec;
	}

	//-------------------------------------------------------------------------

	// Agent implementation
	public Agent split ()
	{
		CLAgent agn = null;

		try{
			agn = (CLAgent)clone();
			agn.mutateGenotype();
		}
		catch ( Exception e ) {
			System.err.println("Fail in split for "+getClass().getName());
			e.printStackTrace();
			System.exit(1);
		}

		return agn;
	}

	//-------------------------------------------------------------------------

	/**
    * Mutates the genetic material of instances.
	 */
	protected void mutateGenotype ()
	{
		int    iAtts = galeCnf.getNumAttributes();
		int     iCls = galeCnf.getNumClasses();
		float    fPM = galeCnf.getSomaticMP();
		int [] iaTmp = null;

		for ( int i=0, iMax=genVec.size() ; i<iMax ; i++ ) {
			iaTmp = (int[])genVec.elementAt(i);
			for ( int j=0 ; j<iAtts ; j++ )
				if ( rnd.nextFloat()<fPM )
					if ( rnd.nextBoolean() )
						iaTmp[j] = 0;
					else
						iaTmp[j] = rnd.nextInt((int)attCnf.attribute(j).upperBound())+1;
			if ( rnd.nextFloat()<fPM )
				iaTmp[iAtts] = rnd.nextInt(iCls);
		}
	}

	//-------------------------------------------------------------------------

	// Agent implementation
	public Object clone()
	throws CloneNotSupportedException
	{
		CLAgent nag = (CLAgent)super.clone();

      nag.galeCnf = galeCnf;
		nag.attCnf  = attCnf;
		nag.rnd     = rnd;
		nag.agnPer  = (AgentPerformance)agnPer.clone();

		nag.genVec = new Vector();
		for ( int i=0, iMax=genVec.size() ; i<iMax ; i++ )
			nag.genVec.addElement(((int[])genVec.elementAt(i)).clone());

		nag.iaPrune = (int[])iaPrune.clone();
		nag.iaMatch = (int[])iaMatch.clone();

      return nag;
	}

	//-------------------------------------------------------------------------

	// Agent implementation
	public float complexity ()
	{
		return genVec.size();
	}

	//-------------------------------------------------------------------------

	// Agent implementation
	public Agent prune ()
	{
		Agent  agn   = null;
		int    iVec  = 0;
		int    iArr  = 0;
		int    iMax  = genVec.size();
		int    iNew  = genVec.size();
		int [] iaTmp = null;
		int [] iaMtc = null;

		for ( ; iArr<iMax ; iArr++ ) {
			if ( iaPrune[iArr]<=0 ) {
				genVec.removeElementAt(iVec);
				iNew--;
			}
			else 
				iVec++;
		}

		if ( genVec.size()==0 ) 
			agn = null;
		else if ( iMax==iNew )
			agn = this;
		else {
			// Remove useless things
			iArr = iVec = 0;
			iaTmp = new int[genVec.size()];
			iaMtc = new int[genVec.size()];
			while ( iArr<iMax && iVec<iaTmp.length ) {
				if ( iaPrune[iArr]>0 ) {
					iaTmp[iVec] = iaPrune[iArr];
					iaMtc[iVec] = iaMatch[iArr];
					iArr++;
					iVec++;
				}
				else
					iArr++;
			}
			iaPrune = iaTmp;
			iaMatch = iaMtc;
			agn = this;
		}

		return agn;
	}

	//-------------------------------------------------------------------------

	// Agent implementation
	public int classify ( Instance ins )
	{
		int      iCls = -1;
		int      iRes = ins.getInstanceClass();
		int      iTmp = 0;
		int []     ia = null;
		boolean bFlag = true;

		for ( int i=0, iMax=genVec.size() ; bFlag && i<iMax ; i++ )
		{
			ia = (int[])genVec.elementAt(i);
			if ( matched(ia,ins) ) {
				iaMatch[i]++;
				bFlag = false;
				iCls = ia[ia.length-1];
				if ( iCls==iRes )
					iaPrune[i]++;
			}
		}
		agnPer.update(iCls,iRes);

		return iCls;
	}

	//-----------------------------------------------------------------------

	/** Check if the rule is matched by the instance.
	 *
	 * @param ia the rule to be check
	 * @param ins the instance
	 * @return true if the rule is matched
	 */
	protected boolean matched ( int [] ia, Instance ins )
	{
		int iAtts = galeCnf.getNumAttributes();
		boolean b = true;

		for ( int i=0 ; b && i<iAtts ; i++ )
			if ( ia[i]!=0 && (ia[i]-1)!=(int)ins.getValue(i) && ins.isKnown(i) )
				b = false;
		
		return b;
	}

	//-------------------------------------------------------------------------

	// Agent implementation
	public String toString()
	{
		String sRes = "Instances of the model\n----------------------\n\n";
		int   [] ia = null;
		int    iAtt = galeCnf.getNumAttributes();

		sRes += "Number of rules : "+genVec.size()+"\n\n";

		for ( int i=0, iMax=genVec.size() ; i<iMax ; i++ ) {
			ia = (int[])genVec.elementAt(i);
			for ( int j=0 ; j<iAtt ; j++ ) {
				if ( ia[j]==0 ) sRes += "#";
				else            sRes += ia[j]-1;
			}
			sRes += ":"+ia[iAtt]+" ("+iaPrune[i]+"/"+iaMatch[i]+")\n";
		}

		sRes += "\n"+agnPer;
			
		return sRes;
	}

	//-------------------------------------------------------------------------

	public static void main ( String [] sArgs )
	{
		try {
			GALEConfiguration gg = new FileGALEConfiguration("gmx3.txt");
			InstanceSet       ii = new FileInstanceSet(gg,"mx3add2.arff");

			gg.getDistance().init(ii.attributeSet());

			Agent a = new CLAgent(gg,ii.attributeSet());

			for ( int i=0, iMax=ii.getSize() ; i<iMax ; i++ ) {
				a.classify(ii.getInstance(i));	
			}

			System.out.println(a);
			a = a.prune();
			System.out.println("Pruned\n==========\n"+a);
			Agent b = a.split();
			System.out.println(b);
			Agent c = a.merge(b);
			System.out.println(c);

		}
		catch ( Exception e ) {
			e.printStackTrace();
		}
	}

}
